
# Originally from:
#     name="requirements-detector",
#     url="https://github.com/landscapeio/requirements-detector",
#     author="landscape.io",
#     author_email="code@landscape.io",
#
# SPDX-License-Identifier: MIT
#
# The MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import os
from unittest import TestCase

from pip_requirements_parser import RequirementsFile


class DependencyDetectionTest(TestCase):

    def test_requirements_txt_parsing(self):
        filepath = os.path.join(os.path.dirname(__file__), "requirements_detector/test1/requirements.txt")
        dependencies = RequirementsFile.from_file(filepath)
        results = [str(r.req) for r in dependencies.requirements]

        expected = ["Django>=1.5.0", "South==0.8.2", "amqp!=1.0.13", "six<1.4,>=1.3.0"]

        assert results == expected
        assert dependencies.options[0].options == {"index_url": "https://example.com/custom/pypi"}
        assert dependencies.comments[0]. line == "# we want six too"

    def test_requirements_dir_parsing1(self):
        filepath = os.path.join(os.path.dirname(__file__), "requirements_detector/test2/requirements/base.txt")
        dependencies = RequirementsFile.from_file(filepath)
        results = [str(r.req) for r in dependencies.requirements]

        expected = [
            "amqp==1.0.13",
            "anyjson==0.3.3",
        ]

        assert results == expected

    def test_requirements_dir_parsing2(self):
        filepath = os.path.join(os.path.dirname(__file__), "requirements_detector/test2/requirements/webui.pip")
        dependencies = RequirementsFile.from_file(filepath)
        results = [str(r.req) for r in dependencies.requirements]

        expected = [
            "Django==1.5.2",
            "South==0.8.2",
        ]

        assert results == expected

    def test_invalid_requirements_txt(self):
        filepath = os.path.join(os.path.dirname(__file__), "requirements_detector/test5/invalid_requirements.txt")
        dependencies = RequirementsFile.from_file(filepath)
        results = [str(r.req) for r in dependencies.requirements]

        expected = ["django<1.6", "django"]
        assert results == expected
        errors = [i.requirement_line.line for i in dependencies.invalid_lines]
        assert errors == ["<<<<<<< HEAD", "=======", ">>>>>>>"]




# Copyright (c) The pip developers (see AUTHORS.txt file)
# SPDX-License-Identifier: MIT

import os
import shutil
import sys
import tempfile
from typing import Tuple

import pytest
from packaging.markers import Marker
from packaging.requirements import Requirement

from pip_requirements_parser import (
    _get_url_from_path,
    _looks_like_path,
    parse_editable,
)
from pip_requirements_parser import build_editable_req
from pip_requirements_parser import build_install_req

from pip_requirements_parser import InstallationError
from pip_requirements_parser import InvalidWheelFilename
from pip_requirements_parser import RequirementLine


class TestInstallRequirement:
    def setup(self) -> None:
        self.tempdir = tempfile.mkdtemp()

    def teardown(self) -> None:
        shutil.rmtree(self.tempdir, ignore_errors=True)

    def test_url_with_query(self) -> None:
        """InstallRequirement should strip the fragment, but not the query."""
        url = "http://foo.com/?p=bar.git;a=snapshot;h=v0.1;sf=tgz"
        fragment = "#egg=bar"
        req = build_install_req(url + fragment)
        assert req.link is not None
        assert req.link.url == url + fragment, req.link

    def test_pep440_wheel_link_requirement(self) -> None:
        line = "test @ https://whatever.com/test-0.4-py2.py3-bogus-any.whl"
        req = build_install_req(line)
        assert str(req.req) == "test@ https://whatever.com/test-0.4-py2.py3-bogus-any.whl"
        assert str(req.link) == "https://whatever.com/test-0.4-py2.py3-bogus-any.whl"

    def test_pep440_url_link_requirement(self) -> None:
        line = "foo @ git+http://foo.com@ref#egg=foo"
        req = build_install_req(line)
        assert str(req.req) == "foo@ git+http://foo.com@ref#egg=foo"
        assert str(req.link) == "git+http://foo.com@ref#egg=foo"

    def test_url_with_authentication_link_requirement(self) -> None:
        url = "https://what@whatever.com/test-0.4-py2.py3-bogus-any.whl"
        line = "https://what@whatever.com/test-0.4-py2.py3-bogus-any.whl"
        req = build_install_req(line)
        assert req.link is not None
        assert req.link.is_wheel
        assert req.link.scheme == "https"
        assert req.link.url == url

    def test_unsupported_wheel_link_requirement_raises(self) -> None:
        req = build_install_req(
            "https://whatever.com/peppercorn-0.4-py2.py3-bogus-any.whl",
        )
        assert req.link is not None
        assert req.link.is_wheel
        assert req.link.scheme == "https"

    def test_unsupported_wheel_local_file_requirement_raises(self) -> None:
        req = build_install_req("simple.dist-0.1-py1-none-invalid.whl")
        assert req.link is not None
        assert req.link.is_wheel
        assert req.link.url == "simple.dist-0.1-py1-none-invalid.whl"

    def test_req_dumps(self) -> None:
        req = build_install_req("simple==0.1")
        assert req.dumps() == "simple==0.1"

    def test_invalid_wheel_requirement_raises(self) -> None:
        with pytest.raises(InvalidWheelFilename):
            build_install_req("invalid.whl")

    def test_wheel_requirement_sets_req_attribute(self) -> None:
        req = build_install_req("simple-0.1-py2.py3-none-any.whl")
        assert isinstance(req.req, Requirement)
        assert str(req.req) == "simple==0.1"

    def test_url_preserved_line_req(self) -> None:
        """Confirm the url is preserved in a non-editable requirement"""
        url = "git+http://foo.com@ref#egg=foo"
        req = build_install_req(url)
        assert req.link is not None
        assert req.link.url == url

    def test_build_editable_req_with_dot_and_extras_preserves_extras_on_dumps(self) -> None:
        url = ".[socks]"
        req = build_editable_req(url)
        assert req.extras == {'socks'}
        assert req.dumps() == "--editable .[socks]"

    def test_build_editable_req_with_dot_dumps(self) -> None:
        url = "."
        req = build_editable_req(url)
        assert req.dumps() == "--editable ."

    def test_url_preserved_editable_req(self) -> None:
        """Confirm the url is preserved in a editable requirement"""
        url = "git+http://foo.com@ref#egg=foo"
        req = build_editable_req(url)
        assert req.link is not None
        assert req.link.url == url

    def test_marker(self) -> None:
        for line in (
            # recommended syntax
            'mock3; python_version >= "3"',
            # with more spaces
            'mock3 ; python_version >= "3" ',
            # without spaces
            'mock3;python_version >= "3"',
        ):
            req = build_install_req(line)
            assert req.req is not None
            assert req.req.name == "mock3"
            assert not req.specifier
            assert str(req.marker) == 'python_version >= "3"'

    def test_marker_semicolon(self) -> None:
        # check that the marker can contain a semicolon
        req = build_install_req('semicolon; os_name == "a; b"')
        assert req.req is not None
        assert req.req.name == "semicolon"
        assert not req.specifier
        assert str(req.marker) == 'os_name == "a; b"'

    def test_marker_url(self) -> None:
        # test "URL; marker" syntax
        url = "http://foo.com/?p=bar.git;a=snapshot;h=v0.1;sf=tgz"
        line = f'{url}; python_version >= "3"'
        req = build_install_req(line)
        assert req.link is not None
        assert req.link.url == url, req.link.url
        assert str(req.marker) == 'python_version >= "3"'

        # without space, marker are part of the URL
        url = "http://foo.com/?p=bar.git;a=snapshot;h=v0.1;sf=tgz"
        line = f'{url};python_version >= "3"'
        req = build_install_req(line)
        assert req.link is not None
        assert req.link.url == line, req.link.url
        assert req.marker is None

    def test_marker_match_from_line(self) -> None:
        # match
        for marker in (
            'python_version >= "1.0"',
            f"sys_platform == {sys.platform!r}",
        ):
            line = "name; " + marker
            req = build_install_req(line)
            assert str(req.marker) == str(Marker(marker))
            assert req.match_marker()

        # don't match
        for marker in (
            'python_version >= "5.0"',
            f"sys_platform != {sys.platform!r}",
        ):
            line = "name; " + marker
            req = build_install_req(line)
            assert str(req.marker) == str(Marker(marker))
            assert not req.match_marker()

    def test_marker_match(self) -> None:
        # match
        for marker in (
            'python_version >= "1.0"',
            f"sys_platform == {sys.platform!r}",
        ):
            line = "name; " + marker
            req = build_install_req(line, requirement_line=None)
            assert str(req.marker) == str(Marker(marker))
            assert req.match_marker()

        # don't match
        for marker in (
            'python_version >= "5.0"',
            f"sys_platform != {sys.platform!r}",
        ):
            line = "name; " + marker
            req = build_install_req(line, requirement_line=None)
            assert str(req.marker) == str(Marker(marker))
            assert not req.match_marker()

    def test_extras_for_line_path_requirement(self) -> None:
        line = "SomeProject[ex1,ex2]"
        filename = "filename"
        requirement_line = RequirementLine(
            filename=filename,
            line_number=1,
            line=line,
        )
        req = build_install_req(line, requirement_line=requirement_line)
        assert len(req.extras) == 2
        assert req.extras == {"ex1", "ex2"}

    def test_extras_for_line_url_requirement(self) -> None:
        line = "git+https://url#egg=SomeProject[ex1,ex2]"
        filename = "filename"
        requirement_line = RequirementLine(
            filename=filename,
            line_number=1,
            line=line,
        )
        req = build_install_req(line, requirement_line=requirement_line)
        assert len(req.extras) == 2
        assert req.extras == {"ex1", "ex2"}

    def test_extras_for_editable_path_requirement(self) -> None:
        url = ".[ex1,ex2]"
        filename = "filename"

        requirement_line = RequirementLine(
            filename=filename,
            line_number=1,
            line=url,
        )
        req = build_editable_req(url, requirement_line=requirement_line)
        assert len(req.extras) == 2
        assert req.extras == {"ex1", "ex2"}

    def test_extras_for_editable_url_requirement(self) -> None:
        url = "git+https://url#egg=SomeProject[ex1,ex2]"
        filename = "filename"
        requirement_line = RequirementLine(
            filename=filename,
            line_number=1,
            line=url,
        )
        req = build_editable_req(url, requirement_line=requirement_line)
        assert len(req.extras) == 2
        assert req.extras == {"ex1", "ex2"}

    def test_unexisting_path(self) -> None:
        result = build_install_req(os.path.join("this", "path", "does", "not", "exist"))
        assert result.link.url == "this/path/does/not/exist"

    def test_single_equal_sign(self) -> None:
        with pytest.raises(InstallationError):
            build_install_req("toto=42")

    def test_unidentifiable_name(self) -> None:
        test_name = "-"
        with pytest.raises(InstallationError):
            build_install_req(test_name)

    def test_requirement_file(self) -> None:
        req_file_path = os.path.join(self.tempdir, "test.txt")
        with open(req_file_path, "w") as req_file:
            req_file.write("pip\nsetuptools")
        result =  build_install_req(req_file_path)
        assert result.req is None
        assert result.link.url.endswith("test.txt")


def test_parse_editable_explicit_vcs() -> None:
    assert parse_editable("svn+https://foo#egg=foo") == (
        "foo",
        "svn+https://foo#egg=foo",
        set(),
    )


def test_parse_editable_with_egg_fragment() -> None:
    assert parse_editable("path/to/AnotherProject#egg=AnotherProject") == (
        "AnotherProject",
        "path/to/AnotherProject#egg=AnotherProject",
        set()
    )

def test_parse_editable_without_egg_fragment() -> None:
    assert parse_editable("path/to/AnotherProject") == (
        None,
        "path/to/AnotherProject",
        set()
    )


def test_parse_editable_example_from_docstring1() -> None:
    # taken from the parse_editable() docstring
    # it is not clear if this is supported and extra are not detected correctly
    # this is a rare edge case anyway
    assert parse_editable("svn+http://blahblah@rev#egg=Foobar[baz]&subdirectory=version_subdir") == (
        'Foobar[baz]',
        'svn+http://blahblah@rev#egg=Foobar[baz]&subdirectory=version_subdir',
        set(),
    )


def test_parse_editable_example_from_docstring2() -> None:
    # taken from the parse_editable() docstring
    assert parse_editable(".[some_extra]") == (None, '.', {'some_extra'})


def test_parse_editable_vcs_extras() -> None:
    assert parse_editable("svn+https://foo#egg=foo[extras]") == (
        "foo",
        "svn+https://foo#egg=foo",
         {'extras'},
    )


def test_exclusive_environment_marker() -> None:
    """Make sure RequirementSet accepts several excluding env marker"""
    eq36 = build_install_req("Django>=1.6.10,<1.7 ; python_version == '3.6'")
    eq36.user_supplied = True
    ne36 = build_install_req("Django>=1.6.10,<1.8 ; python_version != '3.6'")
    ne36.user_supplied = True



@pytest.mark.parametrize(
    "path, expected",
    [
        # Test UNIX-like paths
        ("/path/to/installable", True),
        # Test relative paths
        ("./path/to/installable", True),
        # Test current path
        (".", True),
        # Test url paths
        ("https://whatever.com/test-0.4-py2.py3-bogus-any.whl", True),
        # Test pep440 paths
        ("test @ https://whatever.com/test-0.4-py2.py3-bogus-any.whl", True),
        # Test wheel
        ("simple-0.1-py2.py3-none-any.whl", False),
        # editable egg
        ("path/to/AnotherProject#egg=AnotherProject", True),
    ],
)
def test_looks_like_path(path: str, expected: bool) -> None:
    assert _looks_like_path(path) == expected


@pytest.mark.skipif(
    not sys.platform.startswith("win"), reason="Test only available on Windows"
)
@pytest.mark.parametrize(
    "args, expected",
    [
        # Test relative paths
        ((".\\path\\to\\installable"), True),
        (("relative\\path"), True),
        # Test absolute paths
        (("C:\\absolute\\path"), True),
    ],
)
def test_looks_like_path_win(args: str, expected: bool) -> None:
    assert _looks_like_path(args) == expected


@pytest.mark.parametrize(
    "args, expected",
    [
        # Test pep440 urls
        (
            (
                "/path/to/foo @ git+http://foo.com@ref#egg=foo",
                "foo @ git+http://foo.com@ref#egg=foo",
            ),
            "/path/to/foo @ git+http://foo.com@ref#egg=foo",
        ),
        # Test pep440 urls without spaces
        (
            (
                "/path/to/foo@git+http://foo.com@ref#egg=foo",
                "foo @ git+http://foo.com@ref#egg=foo",
            ),
            "/path/to/foo@git+http://foo.com@ref#egg=foo",
        ),
        # Test pep440 wheel
        (
            (
                "/path/to/test @ https://whatever.com/test-0.4-py2.py3-bogus-any.whl",
                "test @ https://whatever.com/test-0.4-py2.py3-bogus-any.whl",
            ),
            "/path/to/test @ https://whatever.com/test-0.4-py2.py3-bogus-any.whl",
        ),
        # Test name is not a file
        (("/path/to/simple==0.1", "simple==0.1"), None),
    ],
)
def test_get_url_from_path(
    args: Tuple[str, str],
    expected: None,
) -> None:
    assert _get_url_from_path(*args) == expected


def test_get_url_from_path__archive_file() -> None:
    name = "simple-0.1-py2.py3-none-any.whl"
    path = os.path.join("/path/to/" + name)
    assert _get_url_from_path(path, name) == path


def test_get_url_from_path__installable_error() -> None:
    name = "some/setuptools/project"
    path = os.path.join("/path/to/" + name)
    result = _get_url_from_path(path, name)
    assert result == "/path/to/some/setuptools/project"




# Copyright (c) The pip developers (see AUTHORS.txt file)
# SPDX-License-Identifier: MIT

# The following comment should be removed at some point in the future.
# mypy: strict-optional=False

import errno
import logging
import os
import shutil
import stat
from types import TracebackType
from typing import (
    Any,
    Callable,
    Optional,
    Tuple,
    Type,
    TypeVar,
)


logger = logging.getLogger(__name__)

T = TypeVar("T")
ExcInfo = Tuple[Type[BaseException], BaseException, TracebackType]
VersionInfo = Tuple[int, int, int]
NetlocTuple = Tuple[str, Tuple[Optional[str], Optional[str]]]


def ensure_dir(path: str) -> None:
    """os.path.makedirs without EEXIST."""
    try:
        os.makedirs(path)
    except OSError as e:
        # Windows can raise spurious ENOTEMPTY errors. See #6426.
        if e.errno != errno.EEXIST and e.errno != errno.ENOTEMPTY:
            raise


def rmtree(directory: str, ignore_errors: bool = False) -> None:
    shutil.rmtree(directory, ignore_errors=ignore_errors, onerror=rmtree_errorhandler)


def rmtree_errorhandler(func: Callable[..., Any], path: str, exc_info: ExcInfo) -> None:
    """On Windows, the files in .svn are read-only, so when rmtree() tries to
    remove them, an exception is thrown.  We catch that here, remove the
    read-only attribute, and hopefully continue without problems."""
    try:
        has_attr_readonly = not (os.stat(path).st_mode & stat.S_IWRITE)
    except OSError:
        # it's equivalent to os.path.exists
        return

    if has_attr_readonly:
        # convert to read/write
        os.chmod(path, stat.S_IWRITE)
        # use the original function to repeat the operation
        func(path)
        return
    else:
        raise


# Simulates an enum
def enum(*sequential: Any, **named: Any) -> Type[Any]:
    enums = dict(zip(sequential, range(len(sequential))), **named)
    reverse = {value: key for key, value in enums.items()}
    enums["reverse_mapping"] = reverse
    return type("Enum", (), enums)





# Copyright (c) The pip developers (see AUTHORS.txt file)
# SPDX-License-Identifier: MIT

import os
import shutil

from contextlib import contextmanager
from typing import (
    Iterator,
    List,
    Optional,
)

from pip_requirements_parser_tests.lib.path import Path


DATA_DIR = Path(__file__).parent.parent.joinpath("data").resolve()
SRC_DIR = Path(__file__).resolve().parent.parent.parent

REQFILES_DIR = DATA_DIR.joinpath("requirements")

ALL_REQFILES = [
    REQFILES_DIR.joinpath(rf) 
    for rf in os.listdir(REQFILES_DIR)
    if rf.endswith(".txt")
]


MORE_REQFILES_DIR = DATA_DIR.joinpath("more-requirements")

MORE_REQFILES = [
    MORE_REQFILES_DIR.joinpath(rf) 
    for rf in os.listdir(MORE_REQFILES_DIR)
    if rf.endswith(".txt")
]



class TestData:
    """
    Represents a bundle of pre-created test data.

    This copies a pristine set of test data into a root location that is
    designed to be test specific. The reason for this is when running the tests
    concurrently errors can be generated because the related tooling uses
    the directory as a work space. This leads to two concurrent processes
    trampling over each other. This class gets around that by copying all
    data into a directory and operating on the copied data.
    """

    __test__ = False

    def __init__(self, root: str, source: Optional[Path] = None) -> None:
        self.source = source or DATA_DIR
        self.root = Path(root).resolve()

    @classmethod
    def copy(cls, root: str) -> "TestData":
        obj = cls(root)
        obj.reset()
        return obj

    def reset(self) -> None:
        # Check explicitly for the target directory to avoid overly-broad
        # try/except.
        if self.root.exists():
            shutil.rmtree(self.root)
        shutil.copytree(self.source, self.root, symlinks=True)

    @property
    def packages(self) -> Path:
        return self.root.joinpath("packages")

    @property
    def reqfiles(self) -> Path:
        return self.root.joinpath("reqfiles")


@contextmanager
def requirements_file(contents: str, tmpdir: Path) -> Iterator[Path]:
    """Return a Path to a requirements file of given contents.

    As long as the context manager is open, the requirements file will exist.

    :param tmpdir: A Path to the folder in which to create the file

    """
    path = tmpdir / "reqs.txt"
    path.write_text(contents)
    yield path
    path.unlink()




# Copyright (c) The pip developers (see AUTHORS.txt file)
# SPDX-License-Identifier: MIT

import pytest

from pip_requirements_parser import Link
from pip_requirements_parser import links_equivalent

class TestLink:
    @pytest.mark.parametrize(
        "url, expected",
        [
            (
                "https://user:password@example.com/path/page.html",
                "<Link https://user:password@example.com/path/page.html>",
            ),
        ],
    )
    def test_repr(self, url: str, expected: str) -> None:
        link = Link(url)
        assert repr(link) == expected

    @pytest.mark.parametrize(
        "url, expected",
        [
            ("http://yo/wheel.whl", "wheel.whl"),
            ("http://yo/wheel", "wheel"),
            ("https://example.com/path/page.html", "page.html"),
            # Test a quoted character.
            ("https://example.com/path/page%231.html", "page#1.html"),
            (
                "http://yo/myproject-1.0%2Bfoobar.0-py2.py3-none-any.whl",
                "myproject-1.0+foobar.0-py2.py3-none-any.whl",
            ),
            # Test a path that ends in a slash.
            ("https://example.com/path/", "path"),
            ("https://example.com/path//", "path"),
            # Test a url with no filename.
            ("https://example.com/", "example.com"),
            # Test a url with no filename and with auth information.
            (
                "https://user:password@example.com/",
                "example.com",
            ),
        ],
    )
    def test_filename(self, url: str, expected: str) -> None:
        link = Link(url)
        assert link.filename == expected

    def test_splitext(self) -> None:
        assert ("wheel", ".whl") == Link("http://yo/wheel.whl").splitext()

    def test_no_ext(self) -> None:
        assert "" == Link("http://yo/wheel").ext

    def test_ext(self) -> None:
        assert ".whl" == Link("http://yo/wheel.whl").ext

    def test_ext_fragment(self) -> None:
        assert ".whl" == Link("http://yo/wheel.whl#frag").ext

    def test_ext_query(self) -> None:
        assert ".whl" == Link("http://yo/wheel.whl?a=b").ext

    def test_is_wheel(self) -> None:
        assert Link("http://yo/wheel.whl").is_wheel

    def test_is_wheel_false(self) -> None:
        assert not Link("http://yo/not_a_wheel").is_wheel

    def test_fragments(self) -> None:
        url = "git+https://example.com/package#egg=eggname"
        assert "eggname" == Link(url).egg_fragment
        assert None is Link(url).subdirectory_fragment
        url = "git+https://example.com/package#egg=eggname&subdirectory=subdir"
        assert "eggname" == Link(url).egg_fragment
        assert "subdir" == Link(url).subdirectory_fragment
        url = "git+https://example.com/package#subdirectory=subdir&egg=eggname"
        assert "eggname" == Link(url).egg_fragment
        assert "subdir" == Link(url).subdirectory_fragment

    @pytest.mark.parametrize(
        "url, expected",
        [
            ("git+https://github.com/org/repo", True),
            ("bzr+http://bzr.myproject.org/MyProject/trunk/#egg=MyProject", True),
            ("hg+file://hg.company.com/repo", True),
            ("https://example.com/some.whl", False),
            ("file://home/foo/some.whl", False),
        ],
    )
    def test_is_vcs(self, url: str, expected: bool) -> None:
        link = Link(url)
        assert link.is_vcs is expected


@pytest.mark.parametrize(
    "url1, url2",
    [
        pytest.param(
            "https://example.com/foo#egg=foo",
            "https://example.com/foo",
            id="drop-egg",
        ),
        pytest.param(
            "https://example.com/foo#subdirectory=bar&egg=foo",
            "https://example.com/foo#subdirectory=bar&egg=bar",
            id="drop-egg-only",
        ),
        pytest.param(
            "https://example.com/foo#subdirectory=bar&egg=foo",
            "https://example.com/foo#egg=foo&subdirectory=bar",
            id="fragment-ordering",
        ),
        pytest.param(
            "https://example.com/foo?a=1&b=2",
            "https://example.com/foo?b=2&a=1",
            id="query-opordering",
        ),
    ],
)
def test_links_equivalent(url1: str, url2: str) -> None:
    assert links_equivalent(Link(url1), Link(url2))


@pytest.mark.parametrize(
    "url1, url2",
    [
        pytest.param(
            "https://example.com/foo#sha512=1234567890abcdef",
            "https://example.com/foo#sha512=abcdef1234567890",
            id="different-keys",
        ),
        pytest.param(
            "https://example.com/foo#sha512=1234567890abcdef",
            "https://example.com/foo#md5=1234567890abcdef",
            id="different-values",
        ),
        pytest.param(
            "https://example.com/foo#subdirectory=bar&egg=foo",
            "https://example.com/foo#subdirectory=rex",
            id="drop-egg-still-different",
        ),
    ],
)
def test_links_equivalent_false(url1: str, url2: str) -> None:
    assert not links_equivalent(Link(url1), Link(url2))




# Copyright (c) The pip developers (see AUTHORS.txt file)
# SPDX-License-Identifier: MIT

import pathlib
import textwrap
from typing import Callable
from typing import Iterator, List, Union
from unittest import mock

import pytest
from packaging.specifiers import SpecifierSet

import pip_requirements_parser  # this will be monkeypatched
from pip_requirements_parser import RequirementsFileParseError
from pip_requirements_parser import build_editable_req
from pip_requirements_parser import build_install_req
from pip_requirements_parser import (
    break_args_options,
    split_comments,
    join_lines,
    parse_requirements,
    preprocess,
)
from pip_requirements_parser import CommentLine
from pip_requirements_parser import CommentRequirementLine
from pip_requirements_parser import IncorrectRequirementLine
from pip_requirements_parser import InstallRequirement
from pip_requirements_parser import InvalidRequirementLine
from pip_requirements_parser import OptionLine
from pip_requirements_parser import RequirementsFile
from pip_requirements_parser import RequirementLine
from pip_requirements_parser import TextLine

from pip_requirements_parser_tests.lib import TestData, requirements_file
from pip_requirements_parser_tests.lib.path import Path


def get_requirements_and_lines(
    filename: str,
    is_constraint: bool = False,
    include_nested: bool = False,
) -> Iterator[Union[
    InstallRequirement, 
    OptionLine, 
    InvalidRequirementLine, 
    CommentRequirementLine
]]:
    """
    Wrap parse_requirements/build_req_from_parsedreq to
    avoid having to write the same chunk of code in lots of tests.
    """
    return list(RequirementsFile.parse(
        filename=filename,
        is_constraint=is_constraint,
        include_nested=include_nested,
    ))


@pytest.fixture
def parse_requirement_line(
    monkeypatch: pytest.MonkeyPatch,  # NOQA
    tmpdir: Path,
) -> Callable:
    """
    Return a callable to process a single line of text as if it were a full
    requirements file when calling ``parse_requirements``. Writes the line to
    a temp file.
    """

    def process_line(
        line: str,
        filename: str = 'requirements.txt',
        line_number: int = 1,
        is_constraint: bool = False,
    ) -> List[Union[
        InstallRequirement, 
        OptionLine, 
        InvalidRequirementLine, 
        CommentRequirementLine
    ]]:

        prefix = "\n" * (line_number - 1)
        path = tmpdir.joinpath(filename)
        path.parent.mkdir(exist_ok=True)
        path.write_text(prefix + line)
        monkeypatch.chdir(str(tmpdir))
        return get_requirements_and_lines(filename, is_constraint)

    return process_line


@pytest.fixture
def parse_requirement_text(
    monkeypatch: pytest.MonkeyPatch,  # NOQA
    tmpdir: Path,
) -> Callable:
    """
    Return a callable to process some text as if it were requirements file when
    creating a ``RequirementsFile``. Writes the content to a temp file.
    """

    def process_file(
        text: str,
        filename: str = 'requirements.txt',
        include_nested = False,
    ) -> RequirementsFile:
        path = tmpdir.joinpath(filename)
        path.parent.mkdir(exist_ok=True)
        path.write_text(text)
        monkeypatch.chdir(str(tmpdir))
        return RequirementsFile.from_file(path, include_nested=include_nested)

    return process_file


def test_read_file_url(tmp_path: pathlib.Path) -> None:
    reqs = tmp_path.joinpath("requirements.txt")
    reqs.write_text("foo")
    result = list(parse_requirements(reqs.as_posix()))

    assert len(result) == 1
    result = result[0]
    assert result.requirement_string == "foo"
    assert result.requirement_line.to_dict() == dict(line_number=1, line='foo')


class TestPreprocess:
    """tests for `preprocess`"""

    def test_comments_and_joins_case1(self) -> None:
        content = textwrap.dedent(
            """\
          req1 \\
          # comment \\
          req2
        """
        )
        result = preprocess(content)
        assert list(result) == [
            TextLine(line_number=1, line='req1'),
            CommentLine(line_number=1, line='# comment \\'),
            TextLine(line_number=3, line='req2'),
        ]

    def test_comments_and_joins_case2(self) -> None:
        content = textwrap.dedent(
            """\
          req1\\
          # comment
        """
        )
        result = preprocess(content)
        assert list(result) == [
            TextLine(line_number=1, line='req1'),
            CommentLine(line_number=1, line='# comment'),
        ]

    def test_comments_and_joins_case3(self) -> None:
        content = textwrap.dedent(
            """\
          req1 \\
          # comment
          req2
        """
        )
        result = preprocess(content)
        assert list(result) == [
            TextLine(line_number=1, line='req1'),
            CommentLine(line_number=1, line='# comment'),
            TextLine(line_number=3, line='req2'),
        ]


class TestSplitComments:

    def test_split_comments_ignore_empty_line(self) -> None:
        lines = [(1, ""), (2, "req1"), (3, "req2")]
        result = split_comments(lines)
        assert list(result) == [(2, "req1"), (3, "req2")]

    def test_split_comments_returns_comment_line(self) -> None:
        lines = [(1, "req1"), (2, "# comment"), (3, "req2")]
        result = split_comments(lines)
        assert list(result) == [
            TextLine(line_number=1, line='req1'),
            CommentLine(line_number=2, line='# comment'),
            TextLine(line_number=3, line='req2'),
        ]

    def test_split_comments_returns_end_of_line_comment(self) -> None:
        lines = [(1, "req1"), (2, "req # comment"), (3, "req2")]
        result = split_comments(lines)
        assert list(result) == [
            TextLine(line_number=1, line='req1'),
            TextLine(line_number=2, line='req'),
            CommentLine(line_number=2, line='# comment'),
            TextLine(line_number=3, line='req2'),
        ]


class TestJoinLines:
    """tests for `join_lines`"""

    def test_join_lines(self) -> None:
        lines = enumerate(
            [
                "line 1",
                "line 2:1 \\",
                "line 2:2",
                "line 3:1 \\",
                "line 3:2 \\",
                "line 3:3",
                "line 4",
            ],
            start=1,
        )
        expect = [
            (1, "line 1"),
            (2, "line 2:1 line 2:2"),
            (4, "line 3:1 line 3:2 line 3:3"),
            (7, "line 4"),
        ]
        assert expect == list(join_lines(lines))

    def test_last_line_with_escape(self) -> None:
        lines = enumerate(
            [
                "line 1",
                "line 2 \\",
            ],
            start=1,
        )
        expect = [
            (1, "line 1"),
            (2, "line 2 "),
        ]
        assert expect == list(join_lines(lines))


class TestProcessFile:

    def test_can_dumps_name_at_url_requirements_from_file(self, parse_requirement_text) -> None:
        text = "name@http://foo.com"
        rf = parse_requirement_text(text)
        assert rf.dumps() == text

    def test_can_dumps_local_dot_path_from_file(self, parse_requirement_text) -> None:
        text = "."
        rf = parse_requirement_text(text)
        r = rf.requirements[0]
        assert r.name == None
        assert r.link.url == "."

    def test_can_dumps_local_dot_path_extras_from_file(self, parse_requirement_text) -> None:
        text = ". [extra]"
        rf = parse_requirement_text(text)
        r = rf.requirements[0]
        assert r.name == None
        assert r.link.url == "."
        assert r.extras == {"extra"}
        assert r.dumps() == ". [extra]"

    def test_can_dumps_local_path_from_file(self, parse_requirement_text) -> None:
        text = ".foobar"
        rf = parse_requirement_text(text)
        r = rf.requirements[0]
        assert r.name == None
        assert not r.specifier
        assert r.link.url == ".foobar"
        assert r.dumps() == text

    def test_can_dumps_local_path_extras_from_file(self, parse_requirement_text) -> None:
        text = ".foobar [extra]"
        rf = parse_requirement_text(text)
        r = rf.requirements[0]
        assert r.name == None
        assert not r.specifier
        assert r.link.url == ".foobar"
        assert r.extras == {"extra"}
        assert r.dumps() == text

    def test_can_dumps_local_path_to_wheel_from_file(self, parse_requirement_text) -> None:
        text = "./downloads/numpy-1.9.2-cp34-none-win32.whl"
        rf = parse_requirement_text(text)
        r = rf.requirements[0]
        assert r.is_wheel
        assert r.name == "numpy"
        assert r.link.url == "./downloads/numpy-1.9.2-cp34-none-win32.whl"
        assert r.extras == set()
        assert r.dumps() == text

    def test_can_dumps_local_path_to_wheel_with_extras_from_file(self, parse_requirement_text) -> None:
        text = "./downloads/numpy-1.9.2-cp34-none-win32.whl [extra,other]"
        rf = parse_requirement_text(text)
        r = rf.requirements[0]
        assert not r.is_wheel
        assert r.is_local_path
        assert r.name == None
        assert not r.specifier
        assert r.link.url == "./downloads/numpy-1.9.2-cp34-none-win32.whl"
        assert r.extras == {"extra", "other"}
        assert r.dumps() == text

    def test_can_dumps_local_path_to_plain_wheel_from_file(self, parse_requirement_text) -> None:
        text = "numpy-1.9.2-cp34-none-win32.whl"
        rf = parse_requirement_text(text)
        r = rf.requirements[0]
        assert r.is_wheel
        assert r.name == "numpy"
        assert str(r.specifier) == "==1.9.2"
        assert r.link.url == "numpy-1.9.2-cp34-none-win32.whl"
        assert r.extras == set()
        assert r.dumps() == "numpy-1.9.2-cp34-none-win32.whl"

    def test_can_dumps_local_path_to_plain_wheel_with_extras_from_file(self, parse_requirement_text) -> None:
        text = "numpy-1.9.2-cp34-none-win32.whl [extra,other]"
        rf = parse_requirement_text(text)
        r = rf.requirements[0]
        assert r.is_wheel
        assert r.name == "numpy-1.9.2-cp34-none-win32.whl"
        assert not r.specifier
        assert r.link == None
        assert r.extras == {"extra", "other"}
        assert r.dumps() == "numpy-1.9.2-cp34-none-win32.whl [extra,other]"

    def test_can_dumps_local_targz_archive(self, parse_requirement_text) -> None:
        text = "sampleproject-1.0.tar.gz"
        rf = parse_requirement_text(text)
        r = rf.requirements[0]
        assert not r.is_wheel
        assert r.is_archive
        assert r.name == None
        assert not r.specifier
        assert r.link.url == "sampleproject-1.0.tar.gz"
        assert r.extras == set()
        assert r.dumps() == text

    def test_can_dumps_vcs_url_egg_and_extra(self, parse_requirement_text) -> None:
        text = "git+https://github.com/lazzrek/raven-python.git#egg=raven[flask]"
        rf = parse_requirement_text(text)
        r = rf.requirements[0]
        assert not r.is_wheel
        assert not r.is_archive
        assert r.is_vcs_url
        assert r.name == "raven"
        assert not r.specifier
        assert r.link.url == "git+https://github.com/lazzrek/raven-python.git#egg=raven[flask]"
        assert r.extras == {'flask'}
        assert r.dumps() == text

    def test_parse_name_at_url_to_wheel_with_packaging(self, parse_requirement_text) -> None:
        from packaging.requirements import Requirement

        text = "SomeProject@http://my.package.repo/SomeProject2-1.2.3-py33-none-any.whl"
        r = Requirement(text)
        assert r.url =="http://my.package.repo/SomeProject2-1.2.3-py33-none-any.whl"
        assert r.name == "SomeProject"
        assert not r.specifier
        assert r.extras == set()
        assert str(r) == "SomeProject@ http://my.package.repo/SomeProject2-1.2.3-py33-none-any.whl"

    def test_parse_name_at_url__with_packaging(self, parse_requirement_text) -> None:
        from packaging.requirements import Requirement

        text = "SomeProject@http://my.package.repo/SomeProject2.tgz"
        r = Requirement(text)
        assert r.url =="http://my.package.repo/SomeProject2.tgz"
        assert r.name == "SomeProject"
        assert not r.specifier
        assert r.extras == set()
        assert str(r) == "SomeProject@ http://my.package.repo/SomeProject2.tgz"

    def test_parse_name_at_vcs_url_to_wheel_with_packaging(self, parse_requirement_text) -> None:
        from packaging.requirements import Requirement

        text = "SomeProject@git+http://my.package.repo/SomeProject2-1.2.3-py33-none-any.whl"
        r = Requirement(text)
        assert r.url =="git+http://my.package.repo/SomeProject2-1.2.3-py33-none-any.whl"
        assert r.name == "SomeProject"
        assert not r.specifier
        assert r.extras == set()
        assert str(r) == "SomeProject@ git+http://my.package.repo/SomeProject2-1.2.3-py33-none-any.whl"

    def test_can_dumps_name_at_url_to_wheel(self, parse_requirement_text) -> None:
        text = "SomeProject@http://my.package.repo/SomeProject2-1.2.3-py33-none-any.whl"
        rf = parse_requirement_text(text)
        r = rf.requirements[0]
        assert r.is_name_at_url
        assert r.is_wheel
        assert r.is_archive
        assert not r.is_vcs_url
        assert r.name == "SomeProject"
        assert not r.specifier
        assert r.link.url == "http://my.package.repo/SomeProject2-1.2.3-py33-none-any.whl"
        assert r.extras == set()
        assert str(r.req) ==  "SomeProject@ http://my.package.repo/SomeProject2-1.2.3-py33-none-any.whl"
        assert r.dumps() == text

    def test_can_dumps_name_at_url_to_wheel_with_space(self, parse_requirement_text) -> None:
        text = "SomeProject@ http://my.package.repo/SomeProject2-1.2.3-py33-none-any.whl"
        rf = parse_requirement_text(text)
        r = rf.requirements[0]
        assert r.is_name_at_url
        assert r.is_wheel
        assert r.is_archive
        assert not r.is_vcs_url
        assert r.name == "SomeProject"
        assert not r.specifier
        assert r.link.url == "http://my.package.repo/SomeProject2-1.2.3-py33-none-any.whl"
        assert r.extras == set()
        assert str(r.req) == "SomeProject@ http://my.package.repo/SomeProject2-1.2.3-py33-none-any.whl"
        assert r.dumps() == "SomeProject@http://my.package.repo/SomeProject2-1.2.3-py33-none-any.whl"

    def test_can_dumps_name_at_vcs_url_to_wheel(self, parse_requirement_text) -> None:
        text = "SomeProject@git+http://my.package.repo/SomeProject2-1.2.3-py33-none-any.whl"
        rf = parse_requirement_text(text)
        r = rf.requirements[0]
        assert r.is_name_at_url
        assert r.is_wheel
        assert r.is_archive
        assert r.req.url == "git+http://my.package.repo/SomeProject2-1.2.3-py33-none-any.whl"
        assert r.is_vcs_url
        assert r.name == "SomeProject"
        assert not r.specifier
        assert r.link.url == "git+http://my.package.repo/SomeProject2-1.2.3-py33-none-any.whl"
        assert r.extras == set()
        assert r.dumps() == text

    def test_can_dumps_name_at_url_to_wheel_with_space2(self, parse_requirement_text) -> None:
        text = "SomeProject@http://my.package.repo/SomeProject2-1.2.3-py33-none-any.whl"
        rf = parse_requirement_text(text)
        r = rf.requirements[0]
        assert r.is_name_at_url
        assert r.is_wheel
        assert r.is_archive
        assert not r.is_vcs_url
        assert not r.specifier
        assert r.name == "SomeProject"
        assert r.req.name == "SomeProject"
        assert r.link.url == "http://my.package.repo/SomeProject2-1.2.3-py33-none-any.whl"
        assert r.req.url == "http://my.package.repo/SomeProject2-1.2.3-py33-none-any.whl"
        assert r.extras == set()
        assert r.dumps() == text

    def test_can_collect_invalid_top_options_mixed_on_req_line(self, parse_requirement_text) -> None:
        text = "psycopg2>=2.7.4 --no-binary psycopg2"
        rf = parse_requirement_text(text)
        assert not rf.requirements
        assert rf.invalid_lines[0].line == text
        assert rf.invalid_lines[0].error_message==(
            "Invalid global options, not supported with a requirement spec: --no-binary psycopg2"
        )

    def test_can_process_comment_with_colon(self, parse_requirement_text) -> None:
        text = "# No closing quotation: works because we do not validate hashes"
        rf = parse_requirement_text(text)
        co = rf.comments[0]
        assert co.line == "# No closing quotation: works because we do not validate hashes"

    def test_can_process_trailing_quotes(self, parse_requirement_text) -> None:
        text = "FooProject==1.2 --hash=sha256:2cf24dba5fb0a30e26e83b2ac5b9e29e1b161e5c1fa7425e73043362938b9824\","
        rf = parse_requirement_text(text)
        assert not rf.requirements
        il = rf.invalid_lines[0]
        assert il.error_message == "No closing quotation"

    def test_can_dumps_vcs_editable_with_egg_extras_and_other_query_string_args(self, parse_requirement_text) -> None:
        text = "--editable git+https://github.com/techalchemy/pythonfinder.git@master#egg=pythonfinder[dev]&subdirectory=mysubdir"
        rf = parse_requirement_text(text)
        r = rf.requirements[0]
        assert not r.is_name_at_url
        assert r.is_vcs_url
        assert not r.specifier
        assert r.name == "pythonfinder"
        assert r.link.url == "git+https://github.com/techalchemy/pythonfinder.git@master#egg=pythonfinder[dev]&subdirectory=mysubdir"
        assert r.extras == {"dev"}
        assert r.dumps() == text


class TestProcessLine:

    def test_parser_can_detect_editable_egg(self, parse_requirement_line) -> None:
        result = parse_requirement_line(
            "-e path/to/AnotherProject#egg=AnotherProject", "file", 1)
        assert len(result) == 1
        req = result[0]

        requirement_line = RequirementLine(
            line='-e path/to/AnotherProject#egg=AnotherProject',
            line_number=1,
            filename='file',
        )
        assert req.requirement_line == requirement_line
        assert isinstance(req, InstallRequirement)

        assert req.is_editable
        assert req.link.url == "path/to/AnotherProject#egg=AnotherProject" 

    def test_parser_error(self, parse_requirement_line) -> None:
        result = parse_requirement_line("--bogus", "file", 1)
        expected = InvalidRequirementLine(
            requirement_line=RequirementLine(
                line='--bogus',
                line_number=1,
                filename='file',
            ),
            error_message='pip_requirements_parser: error: no such option: --bogus\n',
        )
        assert result == [expected]

    def test_parse_requirement_to_dict(self, parse_requirement_line) -> None:
        line = "pkg==1.0.0 --hash=somehash"
        result = parse_requirement_line(line, "file", 1)[0].to_dict()
        expected = {
            'name': 'pkg',
            'specifier': ['==1.0.0'],
            'extras': [],
            'link': None,
            'marker': None,
            'global_options': [],
            'hash_options': ['somehash'],
            'install_options': [],
            'invalid_options': {},
            
            'is_editable': False,
            'is_constraint': False,
            
            'requirement_line': {
                'line': 'pkg==1.0.0 --hash=somehash',
                'line_number': 1},
            
            'has_egg_fragment': False,
            'is_archive': None,
            'is_local_path': None,
            'is_name_at_url': False,
            'is_pinned': True,
            'is_url': None,
            'is_vcs_url': None,
            'is_wheel': False,
        }
        assert result == expected
 
    def test_parser_non_offending_line(self, parse_requirement_line) -> None:
        try:
            parse_requirement_line("pkg==1.0.0 --hash=sha256:somehash", "file", 1)
        except RequirementsFileParseError:
            pytest.fail("Reported offending line where it should not.")

    def test_only_one_req_per_line(self, parse_requirement_line) -> None:
        reqs = parse_requirement_line(
            line="req1 req2", filename="file", line_number=1
        )
        assert type(reqs[0]) == InvalidRequirementLine

    def test_error_message(self, parse_requirement_line) -> None:
        reqs = parse_requirement_line(
            "my-package=1.0", filename="path/requirements.txt", line_number=3
        )
        assert type(reqs[0]) == InvalidRequirementLine

    def test_yield_line_requirement(self, parse_requirement_line) -> None:
        line = "SomeProject"
        filename = "filename"
        requirement_line = RequirementLine(
            filename=filename,
            line_number=1,
            line=line,
        )
        req = build_install_req(line, requirement_line=requirement_line)
        assert repr(parse_requirement_line(line, filename, 1)[0]) == repr(req)

    def test_yield_pep440_line_requirement(self, parse_requirement_line) -> None:
        line = "SomeProject @ https://url/SomeProject-py2-py3-none-any.whl"
        filename = "filename"
        requirement_line = RequirementLine(
            filename=filename,
            line_number=1,
            line=line,
        )
        req = build_install_req(line, requirement_line=requirement_line)
        assert repr(parse_requirement_line(line, filename, 1)[0]) == repr(req)

    def test_yield_line_constraint(self, parse_requirement_line) -> None:
        line = "SomeProject"
        filename = "filename"
        requirement_line = RequirementLine(
            filename=filename,
            line_number=1,
            line=line,
        )
        req = build_install_req(line, requirement_line=requirement_line, is_constraint=True)
        found_req = parse_requirement_line(line, filename, 1, is_constraint=True)[0]
        assert repr(found_req) == repr(req)
        assert found_req.is_constraint is True

    def test_yield_line_requirement_with_spaces_in_specifier(self, parse_requirement_line) -> None:
        line = "SomeProject >= 2"
        filename = "filename"
        requirement_line = RequirementLine(
            filename=filename,
            line_number=1,
            line=line,
        )
        req = build_install_req(line, requirement_line=requirement_line)
        assert repr(parse_requirement_line(line, filename, 1)[0]) == repr(req)
        assert req.req is not None
        assert str(req.req.specifier) == ">=2"

    def test_yield_editable_requirement(self, parse_requirement_line) -> None:
        url = "git+https://url#egg=SomeProject"
        line = f"-e {url}"
        filename = "filename"
        requirement_line = RequirementLine(
            filename=filename,
            line_number=1,
            line=line,
        )
        req = build_editable_req(url, requirement_line=requirement_line)
        assert repr(parse_requirement_line(line, filename, 1)[0]) == repr(req)

    def test_yield_editable_constraint(self, parse_requirement_line) -> None:
        url = "git+https://url#egg=SomeProject"
        line = f"-e {url}"
        filename = "filename"
        requirement_line = RequirementLine(
            filename=filename,
            line_number=1,
            line=line,
        )
        req = build_editable_req(url, requirement_line=requirement_line, is_constraint=True)
        found_req = parse_requirement_line(line, filename, 1, is_constraint=True)[0]
        assert repr(found_req) == repr(req)
        assert found_req.is_constraint is True

    def test_nested_constraints_file(
        self, 
        monkeypatch: pytest.MonkeyPatch,  # NOQA
        tmpdir: Path,
    ) -> None:

        req_name = "hello"
        req_file = tmpdir / "parent" / "req_file.txt"
        req_file.parent.mkdir()
        req_file.write_text("-c reqs.txt")
        req_file.parent.joinpath("reqs.txt").write_text(req_name)

        monkeypatch.chdir(str(tmpdir))

        reqs = get_requirements_and_lines("./parent/req_file.txt", include_nested=True)
        assert len(reqs) == 2
        assert reqs[0].name == req_name
        assert reqs[0].is_constraint

        assert reqs[1].options ==  {'constraints': ["reqs.txt"]}

    def test_options_on_a_requirement_line(self, parse_requirement_line) -> None:
        line = (
            "SomeProject --install-option=yo1 --install-option yo2 "
            '--global-option="yo3" --global-option "yo4"'
        )
        filename = "filename"
        req = parse_requirement_line(line, filename, 1)[0]
        assert req.global_options == ["yo3", "yo4"]
        assert req.install_options == ["yo1", "yo2"]

    def test_hash_options(self, parse_requirement_line) -> None:
        """Test the --hash option: mostly its value storage.

        Make sure it reads and preserve multiple hashes.

        """
        line = (
            "SomeProject --hash=sha256:2cf24dba5fb0a30e26e83b2ac5b9e29e1b1"
            "61e5c1fa7425e73043362938b9824 "
            "--hash=sha384:59e1748777448c69de6b800d7a33bbfb9ff1b463e44354c"
            "3553bcdb9c666fa90125a3c79f90397bdf5f6a13de828684f "
            "--hash=sha256:486ea46224d1bb4fb680f34f7c9ad96a8f24ec88be73ea8"
            "e5a6c65260e9cb8a7"
        )
        filename = "filename"
        req = parse_requirement_line(line, filename, 1)[0]
        assert sorted(req.hash_options) == [
            "sha256:2cf24dba5fb0a30e26e83b2ac5b9e29e1b161e5c1fa7425e73043362938b9824",
            "sha256:486ea46224d1bb4fb680f34f7c9ad96a8f24ec88be73ea8e5a6c65260e9cb8a7",
            "sha384:59e1748777448c69de6b800d7a33bbfb9ff1b463e44354c3553bcd"
                "b9c666fa90125a3c79f90397bdf5f6a13de828684f"
        ]

    def test_parse_no_index(self, parse_requirement_line) -> None:
        result = parse_requirement_line("--no-index", "file", 1)
        assert result[0].options == {"no_index": True}

    def test_set_finder_index_url(self, parse_requirement_line) -> None:
        result = parse_requirement_line("--index-url=url", "file", 1)
        assert result[0].options == {"index_url": "url"}

    def test_set_finder_find_links(self, parse_requirement_line) -> None:
        result = parse_requirement_line("--find-links=url", "file", 1)
        assert result[0].options == {"find_links": ["url"]}

    def test_set_finder_extra_index_urls(self, parse_requirement_line) -> None:
        result = parse_requirement_line("--extra-index-url=url", "file", 1)
        assert result[0].options == {"extra_index_urls": ["url"]}

    def test_set_finder_allow_all_prereleases(self, parse_requirement_line) -> None:
        result = parse_requirement_line("--pre", "file", 1)
        assert result[0].options == {"pre": True}

    def test_use_feature(self, parse_requirement_line) -> None:
        """--use-feature can be set in requirements files."""
        result = parse_requirement_line("--use-feature=2020-resolver", "file", 1)
        assert result[0].options == {"use_features": ["2020-resolver"]}

    def test_relative_http_nested_req_files(
        self,
        monkeypatch: pytest.MonkeyPatch,  # NOQA
    ) -> None:
        """
        Test a relative nested req file path is joined with the req file url
        """
        req_name = "hello"
        req_file = "http://me.com/me/req_file.txt"

        def get_file_content(filename: str) ->  str:
            if filename == req_file:
                return "-r reqs.txt"
            elif filename == "http://me.com/me/reqs.txt":
                return req_name
            assert False, f"Unexpected file requested {filename}"

        monkeypatch.setattr(
            pip_requirements_parser, "get_file_content", get_file_content
        )

        result = get_requirements_and_lines(req_file, include_nested=True)
        assert len(result) == 2
        assert result[0].name == req_name
        assert not result[0].is_constraint

        assert result[1].options ==  {'requirements': ["reqs.txt"]}

    def test_relative_local_nested_req_files(
        self,
        monkeypatch: pytest.MonkeyPatch,  # NOQA
        tmpdir: Path,
    ) -> None:
        """
        Test a relative nested req file path is joined with the req file dir
        """
        req_name = "hello"
        req_file = tmpdir / "parent" / "req_file.txt"
        req_file.parent.mkdir()
        req_file.write_text("-r reqs.txt")
        req_file.parent.joinpath("reqs.txt").write_text(req_name)

        monkeypatch.chdir(str(tmpdir))

        reqs = get_requirements_and_lines("./parent/req_file.txt", include_nested=True)
        assert len(reqs) == 2
        assert reqs[0].name == req_name
        assert not reqs[0].is_constraint

        assert reqs[1].options ==  {'requirements': ["reqs.txt"]}

    def test_absolute_local_nested_req_files(self, tmpdir: Path) -> None:
        """
        Test an absolute nested req file path
        """
        req_name = "hello"
        req_file = tmpdir / "parent" / "req_file.txt"
        req_file.parent.mkdir()
        other_req_file = tmpdir / "other" / "reqs.txt"
        other_req_file.parent.mkdir()
        # POSIX-ify the path, since Windows backslashes aren't supported.
        other_req_file_str = str(other_req_file).replace("\\", "/")

        req_file.write_text(f"-r {other_req_file_str}")
        other_req_file.write_text(req_name)

        # TODO: also test nested parsing!

        reqs = get_requirements_and_lines(str(req_file), include_nested=True)
        assert len(reqs) == 2
        assert reqs[0].name == req_name
        assert not reqs[0].is_constraint

        assert reqs[1].options ==  {'requirements': [other_req_file_str]}


    def test_absolute_http_nested_req_file_in_local(
        self,
        monkeypatch: pytest.MonkeyPatch,  # NOQA
        tmpdir: Path,
    ) -> None:
        """
        Test a nested req file url in a local req file
        """
        req_name = "hello"
        req_file = tmpdir / "req_file.txt"
        nested_req_file = "http://me.com/me/req_file.txt"

        def get_file_content(filename: str) -> str:
            if filename == str(req_file):
                return f"-r {nested_req_file}"
            elif filename == nested_req_file:
                return req_name

            assert False, f"Unexpected file requested {filename}"

        monkeypatch.setattr(
            pip_requirements_parser, "get_file_content", get_file_content
        )

        result = get_requirements_and_lines(req_file, include_nested=True)
        assert len(result) == 2
        
        assert result[0].name == req_name
        assert not result[0].is_constraint

        assert result[1].options == {'requirements': ['http://me.com/me/req_file.txt']}


class TestBreakOptionsArgs:
    def test_no_args(self) -> None:
        assert ("", "--option") == break_args_options("--option")

    def test_no_options(self) -> None:
        assert ("arg arg", "") == break_args_options("arg arg")

    def test_args_short_options(self) -> None:
        result = break_args_options("arg arg -s")
        assert ("arg arg", "-s") == result

    def test_args_long_options(self) -> None:
        result = break_args_options("arg arg --long")
        assert ("arg arg", "--long") == result


class TestOptionVariants:

    # this suite is really just testing optparse, but added it anyway

    def test_variant1(self, parse_requirement_line) -> None:
        result = parse_requirement_line("-i url", "file", 1)
        assert result[0].options == {"index_url": "url"}

    def test_variant2(self, parse_requirement_line) -> None:
        result = parse_requirement_line("-i 'url'", "file", 1)
        assert result[0].options == {"index_url": "url"}

    def test_variant3(self, parse_requirement_line) -> None:
        result = parse_requirement_line("--index-url=url", "file", 1)
        assert result[0].options == {"index_url": "url"}

    def test_variant4(self, parse_requirement_line) -> None:
        result = parse_requirement_line("--index-url url", "file", 1)
        assert result[0].options == {"index_url": "url"}

    def test_variant5(self, parse_requirement_line) -> None:
        result = parse_requirement_line("--index-url='url'", "file", 1)
        assert result[0].options == {"index_url": "url"}


class TestParseRequirements:
    """tests for `get_requirements_and_lines`"""

    def test_multiple_appending_options(self, tmpdir: Path) -> None:
        with open(tmpdir.joinpath("req1.txt"), "w") as fp:
            fp.write("--extra-index-url url1 \n")
            fp.write("--extra-index-url url2 ")

        
        result = get_requirements_and_lines(tmpdir.joinpath("req1.txt"))
        assert result[0].options =={"extra_index_urls":["url1"]}
        assert result[1].options =={"extra_index_urls":["url2"]}


    def test_expand_missing_env_variables(self, tmpdir: Path) -> None:

        # NOTE: WE DO NOT EXPAND VARS in constrast with PIP
        req_url = (
            "https://${NON_EXISTENT_VARIABLE}:$WRONG_FORMAT@"
            "%WINDOWS_FORMAT%github.com/user/repo/archive/master.zip"
        )

        with open(tmpdir.joinpath("req1.txt"), "w") as fp:
            fp.write(req_url)

        # Construct the session outside the monkey-patch, since it access the
        # env
        with mock.patch("pip_requirements_parser.os.getenv") as getenv:
            getenv.return_value = ""

            reqs = get_requirements_and_lines(tmpdir.joinpath("req1.txt"))

            assert len(reqs) == 1, "parsing requirement file with env variable failed"
            assert reqs[0].link is not None
            assert (
                reqs[0].link.url == req_url
            ), "ignoring invalid env variable in req file failed"

    def test_join_lines(self, tmpdir: Path) -> None:
        with open(tmpdir.joinpath("req1.txt"), "w") as fp:
            fp.write("--extra-index-url url1 \\\n--extra-index-url url2")

        result = get_requirements_and_lines(tmpdir.joinpath("req1.txt"))

        assert result[0].options =={"extra_index_urls":['url1', 'url2']}

    def test_req_file_parse_no_only_binary(self, data: TestData) -> None:
        """
        # default is no constraints
        # We're not testing the format control logic here, just that the options are
        # accepted
        --no-binary fred
        --only-binary wilma
        """
        result = get_requirements_and_lines(
            data.reqfiles.joinpath("supported_options2.txt")
        )
        assert len(result) == 5
        assert all(isinstance(r, CommentRequirementLine) for r in result[:3])
        assert result[3].options == {'no_binary': ['fred']}
        assert result[4].options == {'only_binary': ['wilma']}

    def test_req_file_parse_comment_start_of_line(self, tmpdir: Path) -> None:
        """
        Test parsing comments in a requirements file
        """
        with open(tmpdir.joinpath("req1.txt"), "w") as fp:
            fp.write("# Comment ")

        result = get_requirements_and_lines(tmpdir.joinpath("req1.txt"))
        assert len(result) ==1
        assert isinstance(result[0], CommentRequirementLine)
        assert result[0].line == "# Comment"

    def test_req_file_parse_comment_end_of_line_with_url(self, tmpdir: Path) -> None:
        """
        Test parsing comments in a requirements file
        """
        with open(tmpdir.joinpath("req1.txt"), "w") as fp:
            fp.write("https://example.com/foo.tar.gz # Comment ")

        result = get_requirements_and_lines(tmpdir.joinpath("req1.txt"))
        assert len(result) == 2
        assert result[0].link is not None
        assert result[0].link.url == "https://example.com/foo.tar.gz"
        assert result[1].line == "# Comment"

    def test_req_file_parse_egginfo_end_of_line_with_url(self, tmpdir: Path) -> None:
        """
        Test parsing comments in a requirements file
        """
        with open(tmpdir.joinpath("req1.txt"), "w") as fp:
            fp.write("https://example.com/foo.tar.gz#egg=wat")

        result = get_requirements_and_lines(tmpdir.joinpath("req1.txt"))

        assert len(result) == 1
        assert result[0].name == "wat"

    def test_req_file_no_finder(self, tmpdir: Path) -> None:
        """
        Test parsing a requirements file without a finder
        """
        with open(tmpdir.joinpath("req.txt"), "w") as fp:
            fp.write(
                """
    --find-links https://example.com/
    --index-url https://example.com/
    --extra-index-url https://two.example.com/
    --no-use-wheel
    --no-index
            """
            )

        result = get_requirements_and_lines(tmpdir.joinpath("req.txt"))
        assert len(result) == 6
        assert [type(l) for l in result] == [
            OptionLine,
            OptionLine,
            OptionLine,
            OptionLine,
            IncorrectRequirementLine,
            OptionLine,
        ]

    def test_install_requirements_with_options(
        self,
        tmpdir: Path,
    ) -> None:

        global_option = "--dry-run"
        install_option = "--prefix=/opt"

        content = """
        --only-binary :all:
        INITools==2.0 --global-option="{global_option}" \
                        --install-option "{install_option}"
        """.format(
            global_option=global_option, install_option=install_option
        )

        with requirements_file(content, tmpdir) as reqs_file:
            rf = reqs_file.resolve()
            req = get_requirements_and_lines(rf)

        assert len(req) == 2
        assert req[0].options == {'only_binary': [':all:']}
        assert req[1].name == "INITools"
        assert req[1].specifier == SpecifierSet('==2.0')
        assert req[1].global_options == ["--dry-run"]
        assert req[1].install_options == ["--prefix=/opt"]
